@@ -2,6 +2,7 @@
  *
  *  Copyright (C) 2003 Choe Hwanjin(krisna@kldp.org)
  *  Copyright (c) 2006 Remco den Breeje <remco@sx.mine.nu>
+ *  Copyright (c) 2008 Diego Ongaro <ongardie@gmail.com>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Library General Public License as published
@@ -30,17 +31,47 @@
 #include <libxfcegui4/libxfcegui4.h>
 #include <libxfce4util/libxfce4util.h>
 #include <libxfce4panel/xfce-panel-plugin.h>
+#include <libxfce4panel/xfce-panel-convenience.h>
 
 #include "datetime.h"
 #include "datetime-dialog.h"
 
+#define DATETIME_MAX_STRLEN 256
+
+/**
+ *  Convert a GTimeVal to milliseconds.
+ *  Fractions of a millisecond are truncated.
+ *  With a 32 bit word size and some values of t.tv_sec,
+ *  multiplication by 1000 could overflow a glong,
+ *  so the value is cast to a gint64.
+ */
+static inline gint64 datetime_gtimeval_to_ms(const GTimeVal t)
+{
+  return ((gint64) t.tv_sec * 1000) + ((gint64) t.tv_usec / 1000);
+}
+
+/*
+ * Compute the wake interval,
+ * which is the time remaining from the current time
+ * to the next larger integral multiple of the update interval.
+ * Setting a timer to this value schedules the next update
+ * to occur on the next second or minute
+ * when the given update interval is 1000 or 60000 milliseconds, respectively.
+ */
+static inline guint datetime_wake_interval(const GTimeVal current_time,
+                                           const guint update_interval_ms)
+{
+  return update_interval_ms - (datetime_gtimeval_to_ms(current_time) %
+                               update_interval_ms);
+}
+
 /*
  * Get date/time string
  */
 gchar * datetime_do_utf8strftime(const char *format, const struct tm *tm)
 {
   int len;
-  gchar buf[256];
+  gchar buf[DATETIME_MAX_STRLEN];
   gchar *utf8str = NULL;
 
   /* get formatted date/time */
@@ -56,92 +87,157 @@ gchar * datetime_do_utf8strftime(const char *format, const struct tm *tm)
   return utf8str;
 }
 
+/**
+ *  Check whether date/time format shows seconds
+ */
+static gboolean datetime_format_has_seconds(const gchar *format)
+{
+  static struct tm time_struct = {
+    .tm_sec   = 0,
+    .tm_min   = 0,
+    .tm_hour  = 0,
+    .tm_mday  = 1,
+    .tm_mon   = 0,
+    .tm_year  = 70, /* use 1970 so strftime() can convert '%s' */
+    .tm_wday  = 0,
+    .tm_yday  = 0,
+    .tm_isdst = 0
+  };
+  int len1;
+  int len2;
+  gchar buf1[DATETIME_MAX_STRLEN];
+  gchar buf2[DATETIME_MAX_STRLEN];
+
+  if (format == NULL)
+    return FALSE;
+
+  time_struct.tm_sec = 1;
+  len1 = strftime(buf1, sizeof(buf1)-1, format, &time_struct);
+  if (len1 == 0)
+    return FALSE;
+  buf1[len1] = '\0';
+
+  time_struct.tm_sec = 2;
+  len2 = strftime(buf2, sizeof(buf2)-1, format, &time_struct);
+  if (len2 == 0)
+    return FALSE;
+  buf2[len2] = '\0';
+
+  return len1 != len2 || strcmp(buf1, buf2) != 0;
+}
+
 /*
  * set date and time labels
  */
-gboolean datetime_update(gpointer data)
+gboolean datetime_update(t_datetime *datetime)
 {
   GTimeVal timeval;
   gchar *utf8str;
   struct tm *current;
-  t_datetime *datetime;
+  guint wake_interval;  /* milliseconds to next update */
 
-  if (data == NULL)
+  DBG("wake");
+
+  /* stop timer */
+  if (datetime->timeout_id)
   {
-    return FALSE;
+    g_source_remove(datetime->timeout_id);
   }
 
-  datetime = (t_datetime*)data;
-
   g_get_current_time(&timeval);
   current = localtime((time_t *)&timeval.tv_sec);
-  if (datetime->date_format != NULL && GTK_IS_LABEL(datetime->date_label))
+
+  if (datetime->layout != LAYOUT_TIME &&
+      datetime->date_format != NULL && GTK_IS_LABEL(datetime->date_label))
   {
     utf8str = datetime_do_utf8strftime(datetime->date_format, current);
     gtk_label_set_text(GTK_LABEL(datetime->date_label), utf8str);
     g_free(utf8str);
   }
 
-  if (datetime->time_format != NULL && GTK_IS_LABEL(datetime->time_label))
+  if (datetime->layout != LAYOUT_DATE &&
+      datetime->time_format != NULL && GTK_IS_LABEL(datetime->time_label))
   {
     utf8str = datetime_do_utf8strftime(datetime->time_format, current);
     gtk_label_set_text(GTK_LABEL(datetime->time_label), utf8str);
     g_free(utf8str);
   }
 
-  /* hide labels based on layout-selection */
-  gtk_widget_show(GTK_WIDGET(datetime->time_label));
-  gtk_widget_show(GTK_WIDGET(datetime->date_label));
-  switch(datetime->layout)
-  {
-    case LAYOUT_DATE:
-      gtk_widget_hide(GTK_WIDGET(datetime->time_label));
-      break;
-    case LAYOUT_TIME:
-      gtk_widget_hide(GTK_WIDGET(datetime->date_label));
-      break;
-    default:
-      break;
-  }
+  /* Compute the time to the next update and start the timer. */
+  wake_interval = datetime_wake_interval(timeval, datetime->update_interval);
+  datetime->timeout_id = g_timeout_add(wake_interval, (GSourceFunc) datetime_update, datetime);
 
-  /* set order based on layout-selection */
-  switch(datetime->layout)
-  {
-    case LAYOUT_DATE_TIME:
-      gtk_box_reorder_child(GTK_BOX(datetime->vbox), datetime->time_label, 1);
-      gtk_box_reorder_child(GTK_BOX(datetime->vbox), datetime->date_label, 0);
-      break;
-
-    default:
-      gtk_box_reorder_child(GTK_BOX(datetime->vbox), datetime->time_label, 0);
-      gtk_box_reorder_child(GTK_BOX(datetime->vbox), datetime->date_label, 1);
-  }
   return TRUE;
 }
 
-
-static void on_calendar_entry_activated(GtkWidget *widget, GtkWidget *calendar)
+#if USE_GTK_TOOLTIP_API
+static gboolean datetime_tooltip_timer(t_datetime *datetime)
 {
-  GDate *date;
-  const gchar *text;
+  DBG("wake");
 
-  date = g_date_new();
+  /* flag to datetime_query_tooltip that there is no longer an active timeout */
+  datetime->tooltip_timeout_id = 0;
 
-  text = gtk_entry_get_text(GTK_ENTRY(widget));
-  g_date_set_parse(date, text);
+  /*
+   * Run datetime_query_tooltip if the mouse is still there.
+   * If it is run, it'll register *this* function to be called after another
+   * timeout. If not, *this* function won't run again.
+   */
+  gtk_widget_trigger_tooltip_query(GTK_WIDGET(datetime->button));
 
-  if (g_date_valid(date)) {
-    gtk_calendar_freeze(GTK_CALENDAR(calendar));
-    gtk_calendar_select_month(GTK_CALENDAR(calendar),
-	g_date_get_month(date) - 1,
-	g_date_get_year(date));
-    gtk_calendar_select_day(GTK_CALENDAR(calendar),
-	g_date_get_day(date));
-    gtk_calendar_thaw(GTK_CALENDAR(calendar));
-  }
-  g_date_free(date);
+  /* we don't want to automatically run again */
+  return FALSE;
 }
 
+static gboolean datetime_query_tooltip(GtkWidget *widget,
+                                       gint x, gint y,
+                                       gboolean keyboard_mode,
+                                       GtkTooltip *tooltip,
+                                       t_datetime *datetime)
+{
+  GTimeVal timeval;
+  struct tm *current;
+  gchar *utf8str;
+  gchar *format = NULL;
+  guint wake_interval;  /* milliseconds to next update */
+
+  switch(datetime->layout)
+  {
+    case LAYOUT_TIME:
+      format = datetime->date_format;
+      break;
+    case LAYOUT_DATE:
+      format = datetime->time_format;
+      break;
+    default:
+      break;
+  }
+
+  if (format == NULL)
+    return FALSE;
+
+  g_get_current_time(&timeval);
+  current = localtime((time_t *)&timeval.tv_sec);
+
+  utf8str = datetime_do_utf8strftime(format, current);
+  gtk_tooltip_set_text(tooltip, utf8str);
+  g_free(utf8str);
+
+  /* if there is no active timeout to update the tooltip, register one */
+  if (!datetime->tooltip_timeout_id)
+  {
+    /*
+     * I think we can afford to inefficiently poll every
+     * second while the user keeps the mouse here.
+     */
+    wake_interval = datetime_wake_interval(timeval, 1000);
+    datetime->tooltip_timeout_id = g_timeout_add(wake_interval,
+      (GSourceFunc) datetime_tooltip_timer, datetime);
+  }
+
+  return TRUE;
+}
+#endif
 
 static void on_calendar_realized(GtkWidget *widget, gpointer data)
 {
@@ -176,23 +272,23 @@ static void on_calendar_realized(GtkWidget *widget, gpointer data)
   {
     if (parent_x < root_w / 2) {
       if (parent_y < root_h / 2) {
-	/* upper left */
-	x = parent_x + parent_w;
-	y = parent_y;
+        /* upper left */
+        x = parent_x + parent_w;
+        y = parent_y;
       } else {
-	/* lower left */
-	x = parent_x + parent_w;
-	y = parent_y + parent_h - height;
+        /* lower left */
+        x = parent_x + parent_w;
+        y = parent_y + parent_h - height;
       }
     } else {
       if (parent_y < root_h / 2) {
-	/* upper right */
-	x = parent_x - width;
-	y = parent_y;
+        /* upper right */
+        x = parent_x - width;
+        y = parent_y;
       } else {
-	/* lower right */
-	x = parent_x - width;
-	y = parent_y + parent_h - height;
+        /* lower right */
+        x = parent_x - width;
+        y = parent_y + parent_h - height;
       }
     }
   }
@@ -202,51 +298,55 @@ static void on_calendar_realized(GtkWidget *widget, gpointer data)
     {
       if (parent_y < root_h / 2)
       {
-	/* upper left */
-	x = parent_x;
-	y = parent_y + parent_h;
+        /* upper left */
+        x = parent_x;
+        y = parent_y + parent_h;
       }
       else
       {
-	/* lower left */
-	x = parent_x;
-	y = parent_y - height;
+        /* lower left */
+        x = parent_x;
+        y = parent_y - height;
       }
     }
     else
     {
       if (parent_y < root_h / 2)
       {
-	/* upper right */
-	x = parent_x + parent_w - width;
-	y = parent_y + parent_h;
+        /* upper right */
+        x = parent_x + parent_w - width;
+        y = parent_y + parent_h;
       }
       else
       {
-	/* lower right */
-	x = parent_x + parent_w - width;
-	y = parent_y - height;
+        /* lower right */
+        x = parent_x + parent_w - width;
+        y = parent_y - height;
       }
     }
   }
   gtk_window_move(GTK_WINDOW(widget), x, y);
 }
 
+static gboolean close_calendar_window(t_datetime *datetime)
+{
+  gtk_widget_destroy(datetime->cal);
+  datetime->cal = NULL;
+
+  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(datetime->button), FALSE);
+
+  return TRUE;
+}
 
 /*
  * call the gtk calendar
  */
-static GtkWidget * pop_calendar_window(GtkWidget *parent,
-    int orientation,
-    const char *date_string)
+static GtkWidget * pop_calendar_window(t_datetime *datetime, int orientation)
 {
   GtkWidget *window;
   GtkWidget *frame;
-  GtkWidget *vbox;
-  GtkWidget *hbox;
   GtkWidget *cal;
-  GtkWidget *entry;
-  GtkWidget *label;
+  GtkWidget *parent = datetime->button;
   GdkScreen *screen;
   GtkCalendarDisplayOptions display_options;
   int num;
@@ -255,6 +355,7 @@ static GtkWidget * pop_calendar_window(GtkWidget *parent,
   gtk_window_set_decorated(GTK_WINDOW(window), FALSE);
   gtk_window_set_skip_taskbar_hint(GTK_WINDOW(window), TRUE);
   gtk_window_set_skip_pager_hint(GTK_WINDOW(window), TRUE);
+  gtk_window_stick(GTK_WINDOW(window));
   g_object_set_data(G_OBJECT(window), "calendar-parent", parent);
 
   /* set screen number */
@@ -266,33 +367,23 @@ static GtkWidget * pop_calendar_window(GtkWidget *parent,
   gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_OUT);
   gtk_container_add (GTK_CONTAINER(window), frame);
 
-  vbox = gtk_vbox_new(FALSE, 0);
-  gtk_container_add (GTK_CONTAINER(frame), vbox);
-
   cal = gtk_calendar_new();
   display_options = GTK_CALENDAR_SHOW_HEADING |
     GTK_CALENDAR_SHOW_WEEK_NUMBERS |
     GTK_CALENDAR_SHOW_DAY_NAMES;
   gtk_calendar_display_options(GTK_CALENDAR (cal), display_options);
-  gtk_box_pack_start(GTK_BOX(vbox), cal, TRUE, TRUE, 0);
-
-  hbox = gtk_hbox_new(FALSE, 0);
-  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
-
-  label = gtk_label_new(_("Date:"));
-  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);
-
-  entry = gtk_entry_new();
-  gtk_entry_set_text(GTK_ENTRY(entry), date_string);
-  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
-  g_signal_connect(G_OBJECT(entry), "activate",
-      G_CALLBACK(on_calendar_entry_activated), cal);
+  gtk_container_add (GTK_CONTAINER(frame), cal);
 
   g_signal_connect_after(G_OBJECT(window), "realize",
       G_CALLBACK(on_calendar_realized),
       GINT_TO_POINTER(orientation));
+  g_signal_connect_swapped(G_OBJECT(window), "delete-event",
+      G_CALLBACK(close_calendar_window),
+      datetime);
   gtk_widget_show_all(window);
 
+  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(datetime->button), TRUE);
+
   return window;
 }
 
@@ -305,7 +396,7 @@ static gboolean datetime_clicked(GtkWidget *widget,
 {
   gint orientation;
 
-  if (event->button != 1)
+  if (event->button != 1 || event->state & GDK_CONTROL_MASK)
     return FALSE;
 
   if (datetime == NULL)
@@ -313,8 +404,7 @@ static gboolean datetime_clicked(GtkWidget *widget,
 
   if (datetime->cal != NULL)
   {
-    gtk_widget_destroy(datetime->cal);
-    datetime->cal = NULL;
+    close_calendar_window(datetime);
   }
   else
   {
@@ -322,9 +412,8 @@ static gboolean datetime_clicked(GtkWidget *widget,
     orientation = xfce_panel_plugin_get_orientation(datetime->plugin);
 
     /* draw calendar */
-    datetime->cal = pop_calendar_window(datetime->eventbox,
-	                                orientation,
-	                                gtk_label_get_text(GTK_LABEL(datetime->date_label)));
+    datetime->cal = pop_calendar_window(datetime,
+                                        orientation);
   }
   return TRUE;
 }
@@ -353,6 +442,31 @@ static void datetime_update_time_font(t_datetime *datetime)
   }
 }
 
+static void datetime_set_update_interval(t_datetime *datetime)
+{
+  /* a custom date format could specify seconds */
+  gboolean date_has_seconds = datetime_format_has_seconds(datetime->date_format);
+  gboolean time_has_seconds = datetime_format_has_seconds(datetime->time_format);
+  gboolean has_seconds;
+
+  /* set update interval for the date/time displayed in the panel */
+  switch(datetime->layout)
+  {
+    case LAYOUT_DATE:
+      has_seconds = date_has_seconds;
+      break;
+    case LAYOUT_TIME:
+      has_seconds = time_has_seconds;
+      break;
+    default:
+      has_seconds = date_has_seconds || time_has_seconds;
+      break;
+  }
+
+  /* 1000 ms in 1 second */
+  datetime->update_interval = 1000 * (has_seconds ? 1 : 60);
+}
+
 /*
  * set layout after doing some checks
  */
@@ -362,6 +476,59 @@ void datetime_apply_layout(t_datetime *datetime, t_layout layout)
   {
     datetime->layout = layout;
   }
+
+  /* hide labels based on layout-selection */
+  gtk_widget_show(GTK_WIDGET(datetime->time_label));
+  gtk_widget_show(GTK_WIDGET(datetime->date_label));
+  switch(datetime->layout)
+  {
+    case LAYOUT_DATE:
+      gtk_widget_hide(GTK_WIDGET(datetime->time_label));
+      break;
+    case LAYOUT_TIME:
+      gtk_widget_hide(GTK_WIDGET(datetime->date_label));
+      break;
+    default:
+      break;
+  }
+
+#if USE_GTK_TOOLTIP_API
+  /* update tooltip handler */
+  if (datetime->tooltip_handler_id)
+  {
+    g_signal_handler_disconnect(datetime->button,
+                                datetime->tooltip_handler_id);
+    datetime->tooltip_handler_id = 0;
+  }
+  switch(datetime->layout)
+  {
+    case LAYOUT_DATE:
+    case LAYOUT_TIME:
+      gtk_widget_set_has_tooltip(GTK_WIDGET(datetime->button), TRUE);
+      datetime->tooltip_handler_id = g_signal_connect(datetime->button,
+                             "query-tooltip",
+                             G_CALLBACK(datetime_query_tooltip), datetime);
+      break;
+
+    default:
+      gtk_widget_set_has_tooltip(GTK_WIDGET(datetime->button), FALSE);
+  }
+#endif
+
+  /* set order based on layout-selection */
+  switch(datetime->layout)
+  {
+    case LAYOUT_TIME_DATE:
+      gtk_box_reorder_child(GTK_BOX(datetime->vbox), datetime->time_label, 0);
+      gtk_box_reorder_child(GTK_BOX(datetime->vbox), datetime->date_label, 1);
+      break;
+
+    default:
+      gtk_box_reorder_child(GTK_BOX(datetime->vbox), datetime->time_label, 1);
+      gtk_box_reorder_child(GTK_BOX(datetime->vbox), datetime->date_label, 0);
+  }
+
+  datetime_set_update_interval(datetime);
 }
 
 /*
@@ -400,57 +567,24 @@ void datetime_apply_format(t_datetime *datetime,
   {
     g_free(datetime->date_format);
     datetime->date_format = g_strdup(date_format);
-    if (strlen(date_format) == 0)
-      gtk_widget_hide(GTK_WIDGET(datetime->date_label));
-    else
-      gtk_widget_show(GTK_WIDGET(datetime->date_label));
   }
 
   if (time_format != NULL)
   {
     g_free(datetime->time_format);
     datetime->time_format = g_strdup(time_format);
-    if (strlen(time_format) == 0)
-      gtk_widget_hide(GTK_WIDGET(datetime->time_label));
-    else
-      gtk_widget_show(GTK_WIDGET(datetime->time_label));
   }
 
-  if (datetime->timeout_id)
-  {
-    g_source_remove(datetime->timeout_id);
-  }
-
-  if (strstr(datetime->date_format, "%S") != NULL ||
-      strstr(datetime->date_format, "%s") != NULL ||
-      strstr(datetime->date_format, "%r") != NULL ||
-      strstr(datetime->date_format, "%T") != NULL ||
-      strstr(datetime->time_format, "%S") != NULL ||
-      strstr(datetime->time_format, "%s") != NULL ||
-      strstr(datetime->time_format, "%r") != NULL ||
-      strstr(datetime->time_format, "%T") != NULL)
-  {
-    datetime->timeout_id = g_timeout_add(1000, datetime_update, datetime);
-  }
-  else
-  {
-    datetime->timeout_id = g_timeout_add(10000, datetime_update, datetime);
-  }
+  datetime_set_update_interval(datetime);
 }
 
 /*
- * Set a border - Function only called by the signal handler.
- * A border is only set, if there is enough space for it (size > 26)
+ * Function only called by the signal handler.
  */
 static int datetime_set_size(XfcePanelPlugin *plugin,
     gint size,
     t_datetime *datetime)
 {
-  if(size > 26)
-    gtk_container_set_border_width(GTK_CONTAINER(datetime->frame), 2);
-  else
-    gtk_container_set_border_width(GTK_CONTAINER(datetime->frame), 0);
-
   /* return true to please the signal handler ;) */
   return TRUE;
 }
@@ -464,14 +598,13 @@ static void datetime_read_rc_file(XfcePanelPlugin *plugin, t_datetime *dt)
   XfceRc *rc;
   t_layout layout;
   const gchar *date_font, *time_font, *date_format, *time_format;
-  const gchar *val;
 
   /* load defaults */
   layout = LAYOUT_DATE_TIME;
-  date_font = g_strdup("Bitstream Vera Sans 8");
-  time_font = g_strdup("Bitstream Vera Sans 10");
-  date_format = g_strdup("%Y/%m/%d");
-  time_format = g_strdup("%H:%M");
+  date_font = "Bitstream Vera Sans 8";
+  time_font = "Bitstream Vera Sans 10";
+  date_format = "%Y-%m-%d";
+  time_format = "%H:%M";
 
   /* open file */
   if((file = xfce_panel_plugin_lookup_rc_file(plugin)) != NULL)
@@ -481,20 +614,21 @@ static void datetime_read_rc_file(XfcePanelPlugin *plugin, t_datetime *dt)
 
     if(rc != NULL)
     {
-      layout	  = xfce_rc_read_int_entry(rc, "layout", layout);
-      val = xfce_rc_read_entry(rc, "date_font", date_font);
-      date_font	  = g_strdup(val);
-      val = xfce_rc_read_entry(rc, "time_font", time_font);
-      time_font	  = g_strdup(val);
-      val = xfce_rc_read_entry(rc, "date_format", date_format);
-      date_format = g_strdup(val);
-      val = xfce_rc_read_entry(rc, "time_format", time_format);
-      time_format = g_strdup(val);
+      layout      = xfce_rc_read_int_entry(rc, "layout", layout);
+      date_font   = xfce_rc_read_entry(rc, "date_font", date_font);
+      time_font   = xfce_rc_read_entry(rc, "time_font", time_font);
+      date_format = xfce_rc_read_entry(rc, "date_format", date_format);
+      time_format = xfce_rc_read_entry(rc, "time_format", time_format);
 
       xfce_rc_close(rc);
     }
   }
 
+  date_font   = g_strdup(date_font);
+  time_font   = g_strdup(time_font);
+  date_format = g_strdup(date_format);
+  time_format = g_strdup(time_format);
+
   /* set values in dt struct */
   datetime_apply_layout(dt, layout);
   datetime_apply_font(dt, date_font, time_font);
@@ -515,13 +649,17 @@ void datetime_write_rc_file(XfcePanelPlugin *plugin, t_datetime *dt)
   rc = xfce_rc_simple_open(file, FALSE);
   g_free(file);
 
-  xfce_rc_write_int_entry(rc, "layout", dt->layout);
-  xfce_rc_write_entry(rc, "date_font", dt->date_font);
-  xfce_rc_write_entry(rc, "time_font", dt->time_font);
-  xfce_rc_write_entry(rc, "date_format", dt->date_format);
-  xfce_rc_write_entry(rc, "time_format", dt->time_format);
+  if(rc != NULL)
+  {
+    xfce_rc_write_int_entry(rc, "layout", dt->layout);
+    xfce_rc_write_entry(rc, "date_font", dt->date_font);
+    xfce_rc_write_entry(rc, "time_font", dt->time_font);
+    xfce_rc_write_entry(rc, "date_format", dt->date_format);
+    xfce_rc_write_entry(rc, "time_format", dt->time_format);
+
+    xfce_rc_close(rc);
+  }
 
-  xfce_rc_close(rc);
 }
 
 /*
@@ -529,17 +667,14 @@ void datetime_write_rc_file(XfcePanelPlugin *plugin, t_datetime *dt)
  */
 static void datetime_create_widget(t_datetime * datetime)
 {
-  /* create event box */
-  datetime->eventbox = gtk_event_box_new();
-
-  /* create frame */
-  datetime->frame = gtk_frame_new(NULL);
-  gtk_frame_set_shadow_type(GTK_FRAME(datetime->frame), GTK_SHADOW_NONE);
-  gtk_container_add(GTK_CONTAINER(datetime->eventbox), datetime->frame);
+  /* create button */
+  datetime->button = xfce_create_panel_toggle_button();
+  gtk_widget_show(datetime->button);
 
   /* create vertical box */
   datetime->vbox = gtk_vbox_new(TRUE, 0);
-  gtk_container_add(GTK_CONTAINER(datetime->frame), datetime->vbox);
+  gtk_widget_show(datetime->vbox);
+  gtk_container_add(GTK_CONTAINER(datetime->button), datetime->vbox);
 
   /* create time and date lines */
   datetime->time_label = gtk_label_new("");
@@ -556,7 +691,7 @@ static void datetime_create_widget(t_datetime * datetime)
   gtk_box_reorder_child(GTK_BOX(datetime->vbox), datetime->date_label, 1);
 
   /* connect widget signals to functions */
-  g_signal_connect(datetime->eventbox, "button-press-event",
+  g_signal_connect(datetime->button, "button-press-event",
       G_CALLBACK(datetime_clicked), datetime);
 }
 
@@ -569,31 +704,21 @@ static t_datetime * datetime_new(XfcePanelPlugin *plugin)
 
   DBG("Starting datetime panel plugin");
 
-  /* alloc mem */
+  /* alloc and clear mem */
   datetime = panel_slice_new0 (t_datetime);
 
-  /* set variables */
+  /* store plugin reference */
   datetime->plugin = plugin;
-  datetime->date_font = NULL;
-  datetime->date_format = NULL;
-  datetime->time_font = NULL;
-  datetime->time_format = NULL;
 
   /* call widget-create function */
   datetime_create_widget(datetime);
 
-  /* set calendar variables */
-  datetime->cal = NULL;
-
   /* load settings (default values if non-av) */
   datetime_read_rc_file(plugin, datetime);
 
   /* set date and time labels */
   datetime_update(datetime);
 
-  /* display plugin */
-  gtk_widget_show_all(datetime->eventbox);
-
   return datetime;
 }
 
@@ -602,11 +727,16 @@ static t_datetime * datetime_new(XfcePanelPlugin *plugin)
  */
 static void datetime_free(XfcePanelPlugin *plugin, t_datetime *datetime)
 {
-  /* stop timeout */
-  g_source_remove(datetime->timeout_id);
+  /* stop timeouts */
+  if (datetime->timeout_id != 0)
+    g_source_remove(datetime->timeout_id);
+#if USE_GTK_TOOLTIP_API
+  if (datetime->tooltip_timeout_id != 0)
+    g_source_remove(datetime->tooltip_timeout_id);
+#endif
 
   /* destroy widget */
-  gtk_widget_destroy(datetime->eventbox);
+  gtk_widget_destroy(datetime->button);
 
   /* cleanup */
   g_free(datetime->date_font);
@@ -626,8 +756,8 @@ static void datetime_construct(XfcePanelPlugin *plugin)
   t_datetime * datetime = datetime_new(plugin);
 
   /* add plugin to panel */
-  gtk_container_add(GTK_CONTAINER(plugin), datetime->eventbox);
-  xfce_panel_plugin_add_action_widget(plugin, datetime->eventbox);
+  gtk_container_add(GTK_CONTAINER(plugin), datetime->button);
+  xfce_panel_plugin_add_action_widget(plugin, datetime->button);
 
   /* connect plugin signals to functions */
   g_signal_connect(plugin, "save",
